# -*- coding: utf-8 -*-
import sys
import signal
import pkg_resources
from .styles import CSSParser
from .inputs import StringInput
from .inputs import IntegerInput
from .inputs import SelectInput
from .inputs import BooleanInput
from .inputs import MultiSelectInput
from .inputs import Branch
from .utils import numeric_options


class AddAction(object):

    def __init__(self, form):
        self.form = form

    def __call__(self, key, obj):
        self.form._fields.append((key, obj))

    def string(self, key, label, **kwargs):
        obj = StringInput(label, **kwargs)
        self.form._add(key, obj)
        return self.form

    def int(self, key, label, **kwargs):
        obj = IntegerInput(label, **kwargs)
        self.form._add(key, obj)
        return self.form

    def select(self, key, label, choices, option_format=numeric_options, **kwargs):
        obj = SelectInput(label, choices, option_format, **kwargs)
        self.form._add(key, obj)
        return self.form

    def multiselect(self, key, label, choices, done_label='Done', option_format=numeric_options, **kwargs):
        obj = MultiSelectInput(label, choices, done_label, option_format, **kwargs)
        self.form._add(key, obj)
        return self.form

    def bool(self, key, label, **kwargs):
        obj = BooleanInput(label, **kwargs)
        self.form._add(key, obj)
        return self.form

    def branch(self, handler, *args, **kwargs):
        obj = Branch(handler, *args, **kwargs)
        self.form._add(id(obj), obj)
        return self.form


class Form(object):

    @property
    def add(self):
        action = AddAction(self)
        return action

    def __init__(self):
        self._fields = []
        signal.signal(signal.SIGINT, self._on_sigint)

    def _add(self, key, obj):
        self._fields.append((key, obj))

    def run(self, prefix=None, stylesheet=None):
        styles = None

        if stylesheet:
            styles = CSSParser.parse_string(stylesheet)
        else:
            stream = pkg_resources.resource_stream(
                'promptly.resources',
                'default.css'
            )

            styles = CSSParser.parse_string(stream.read())

        for label, input in iter(self._fields):
            input(form=self, prefix=prefix, stylesheet=styles)

    def __len__(self):
        return len(self._fields)

    def __iter__(self):
        for k, v in iter(self._fields):
            if not isinstance(v, Branch):
                yield k, v.value

    def __getattr__(self, key):
        try:
            # we were using an ordereddict so access was pretty close to
            # O(1), switched to lists in order to facilitate branching
            # this lookup will be O(N)
            return next(x[1] for x in self._fields if x[0] == key)
        except StopIteration:
            raise AttributeError

    # signal handlers
    def _on_sigint(self, signal, frame):
        print ('You Quit! You\'re a quitter! Boo!')
        sys.exit(0)



from .form import Form
from .inputs import StringInput
from .inputs import IntegerInput
from .inputs import SelectInput
from .inputs import MultiSelectInput
from .inputs import BooleanInput
from .inputs import Branch



import sys


_ver = sys.version_info
is_py2 = (_ver[0] == 2)


def iteritems(value):
    if is_py2:
        return value.iteritems()

    return value.items()


def itervalues(value):
    if is_py2:
        return value.itervalues()

    return value.values()


def input(prompt):
    if is_py2:
        return raw_input(prompt)
    else:
        return input(prompt)



# -*- coding: utf-8 -*-
import sys
import signal
import pkg_resources
from .styles import CSSParser
from .compat import iterkeys, input


class BaseRunner(object):

    def __init__(self, prefix=None, stylesheet=None):
        self.prefix = prefix
        self.stylesheet = stylesheet
        signal.signal(signal.SIGINT, self._on_sigint)

    # signal handlers
    def _on_sigint(self, signal, frame):
        print('You Quit! You\'re a quitter! Boo!')
        sys.exit(0)

    def goto(self, key):
        form = self.form
        keys = list(iterkeys(form._fields))

        try:
            index = keys.index(key)
        except ValueError:
            return

        self._sequence = (x for i, x in enumerate(keys) if i > index)

    def info(self, value):
        pass

    def error(self, value):
        pass

    def __call__(self, form):

        self._sequence = iterkeys(form._fields)

        self.form = form
        prefix = '' if not self.prefix else self.prefix

        if self.stylesheet:
            styles = CSSParser.parse_string(self.stylesheet)
        else:
            stream = pkg_resources.resource_stream(
                'promptly.resources',
                'default.css'
            )

            styles = CSSParser.parse_string(stream.read())

        for item in self._sequence:
            target = getattr(form, item)

            prompt = target.build_prompt(prefix=prefix,
                                         stylesheet=styles)

            while True:
                data = self.prompt(prompt, target.)

                try:
                    target(data=data, runner=self)
                except ValueError:
                    continue
                break


class ConsoleRunner(BaseRunner):
    def prompt(self, prompt, ):
        # we want a trailing space
        if
        return input('%s ' % prompt)



# -*- coding: utf-8 -*-
import sys
from .compat import input
from .styles import Style
from .utils import numeric_options


class Branch(object):
    def __init__(self, handler, *args, **kwargs):
        self.handler = handler
        self.args = args
        self.kwargs = kwargs

    def __call__(self, form, prefix=None, stylesheet=None):
        branch = self.handler(form, *self.args, **self.kwargs)

        if not branch:
            return

        index = form._fields.index((id(self), self)) + 1

        for each in iter(branch._fields):
            form._fields.insert(index, each)
            index = index + 1


class BaseInput(object):

    def __init__(self, label, validators=None, default=None, prefix=None):
        self.label = label
        self.default = default
        self.validators = validators
        self.value = None

    def styles_for_key(self, key, stylesheet):
        styles = {}
        context = stylesheet

        try:
            styles.update(context['selectors']['body']['value'])
        except KeyError:
            pass

        for part in key.split('.'):
            try:
                context = context['selectors'][part]
                styles.update(context['value'])

            except KeyError:
                break

        return Style(styles)

    def build_prompt(self, stylesheet=None):
        prompt = self.label

        if self.default:
            prompt = '%s [%s]' % (prompt, self.default)

        return prompt

    def apply_default(self, data):
        if len(data) == 0 and self.default:
            return self.default

        return data

    def process_data(self, data):
        self.value = data

    def validate(self, input):
        return True

    def __call__(self, form, prefix=None, stylesheet=None):
        prompt = '%s' % (self.build_prompt(
            prefix=prefix,
            stylesheet=stylesheet))

        while 1:
            data = self.apply_default(input('%s ' % prompt))

            try:
                self.process_data(data)
            except:
                continue

            if self.validate(self.value):
                break
            else:
                sys.stdout.write('Sorry invalid input\n')
                continue


class StringInput(BaseInput):

    def build_prompt(self, prefix, stylesheet):
        styles_prefix = self.styles_for_key('prefix', stylesheet)
        styles_label = self.styles_for_key('string.label', stylesheet)
        styles_default_wrapper = self.styles_for_key('string.default_wrapper', stylesheet)
        styles_default_value = self.styles_for_key('string.default_value', stylesheet)

        if self.default:
            return '%s%s %s%s%s' % (
                styles_prefix(prefix),
                styles_label(self.label),
                styles_default_wrapper('['),
                styles_default_value(self.default),
                styles_default_wrapper(']'))

        return '%s%s' % (styles_prefix(prefix),
                         styles_label(self.label))


class IntegerInput(BaseInput):

    def __init__(self, label, validators=None, **kwargs):
        super(IntegerInput, self).__init__(label, validators, **kwargs)

    def build_prompt(self, prefix, stylesheet):
        styles_prefix = self.styles_for_key('prefix', stylesheet)
        styles_label = self.styles_for_key('integer.label', stylesheet)
        styles_default_wrapper = self.styles_for_key('integer.default_wrapper', stylesheet)
        styles_default_value = self.styles_for_key('integer.default_value', stylesheet)

        if self.default:
            return '%s%s %s%s%s' % (
                styles_prefix(prefix),
                styles_label(self.label),
                styles_default_wrapper('['),
                styles_default_value(self.default),
                styles_default_wrapper(']'))

        return '%s%s' % (styles_prefix(prefix),
                         styles_label(self.label))

    def process_data(self, data):
        try:
            self.value = int(data)
        except ValueError:
            self.value = None
            raise


class BooleanInput(BaseInput):
    def __init__(self, label, default=True, **kwargs):
        super(BooleanInput, self).__init__(label, default=default, **kwargs)

    def build_prompt(self, prefix, stylesheet):
        styles_prefix = self.styles_for_key('prefix', stylesheet)
        styles_label = self.styles_for_key('boolean.label', stylesheet)
        styles_default_wrapper = self.styles_for_key('boolean.default_wrapper', stylesheet)
        styles_default_value = self.styles_for_key('boolean.default_value', stylesheet)
        styles_other_value = self.styles_for_key('boolean.other_value', stylesheet)
        styles_seperator = self.styles_for_key('boolean.seperator', stylesheet)

        prompt = styles_label(self.label)

        if self.default:
            prompt = '%s%s %s%s%s%s%s' % (
                styles_prefix(prefix),
                prompt,
                styles_default_wrapper('['),
                styles_default_value('Y'),
                styles_seperator('/'),
                styles_other_value('n'),
                styles_default_wrapper(']'))
        else:
            prompt = '%s%s %s%s%s%s%s' % (
                styles_prefix(prefix),
                prompt,
                styles_default_wrapper('['),
                styles_other_value('y'),
                styles_seperator('/'),
                styles_default_value('N'),
                styles_default_wrapper(']'))

        return prompt

    def apply_default(self, data):
        if len(data) == 0:
            return self.default

        return data

    def process_data(self, data):
        boolean_states = {
        True: True, '1': True, 'yes': True, 'true': True, 'on': True, 'y': True, 't': True,
        False: False, '0': False, 'no': False, 'false': False, 'off': False, 'n': False, 'f': False
        }

        candidate = data.lower() if type(data) == str else data

        try:
            self.value = boolean_states[candidate]
        except KeyError:
            raise


class SelectInput(BaseInput):

    def __init__(self, label, choices, option_format=numeric_options, **kwargs):
        super(SelectInput, self).__init__(label, **kwargs)
        self.choices = option_format(choices)

    def build_prompt(self, prefix, stylesheet):
        styles_prefix = self.styles_for_key('prefix', stylesheet)
        styles_label = self.styles_for_key('choices.label', stylesheet)
        styles_default_wrapper = self.styles_for_key('choices.default_wrapper', stylesheet)
        styles_default_value = self.styles_for_key('choices.default_value', stylesheet)
        styles_option_key = self.styles_for_key('choices.option_key', stylesheet)
        styles_option_value = self.styles_for_key('choices.option_value', stylesheet)
        styles_seperator = self.styles_for_key('choices.seperator', stylesheet)
        styles_action = self.styles_for_key('choices.action', stylesheet)

        prompt = '%s\n' % styles_label(self.label)

        choices = []
        for key, value in self.choices:
            choices.append('%s%s %s' % (
                styles_option_key(key),
                styles_seperator(')'),
                styles_option_value(value)))

        prompt = '%s%s\n%s%s' % (
            prompt,
            '\n'.join(choices),
            styles_prefix(prefix),
            styles_action('Select Option'))

        if self.default:
            prompt = '%s %s%s%s' % (prompt,
                styles_default_wrapper('['),
                styles_default_value(self.default),
                styles_default_wrapper(']'))

        return prompt

    def process_data(self, data):
        result = [x for x in self.choices if str(x[0]) == str(data)]

        if not result:
            raise ValueError

        self.value = result[0]


class MultiSelectInput(SelectInput):

    def __init__(self, label, choices, done_label='Done', option_format=numeric_options, **kwargs):
        options = list(choices)
        options += (done_label, )
        super(MultiSelectInput, self).__init__(label, options, option_format, **kwargs)
        self.value = set()

    def build_prompt(self, prefix, stylesheet):
        styles_prefix = self.styles_for_key('prefix', stylesheet)
        styles_label = self.styles_for_key('choices.label', stylesheet)
        styles_default_wrapper = self.styles_for_key('choices.default_wrapper', stylesheet)
        styles_default_value = self.styles_for_key('choices.default_value', stylesheet)
        styles_option_key = self.styles_for_key('choices.option_key', stylesheet)
        styles_option_value = self.styles_for_key('choices.option_value', stylesheet)
        styles_seperator = self.styles_for_key('choices.seperator', stylesheet)
        styles_action = self.styles_for_key('choices.action', stylesheet)

        prompt = '%s\n' % styles_label(self.label)

        choices = []
        choices_count = len(self.choices)

        for i, each in enumerate(self.choices):
            key, value = each
            prefix = '[ ] '

            if i == (choices_count - 1):
                prefix = ''
            elif each in self.value:
                prefix = '[x] '

            choices.append('%s%s %s' % (
                styles_option_key(key),
                styles_seperator(')'),
                styles_option_value('%s%s' % (prefix, value))))

        prompt = '%s%s\n%s%s' % (
            prompt,
            '\n'.join(choices),
            styles_prefix(prefix),
            styles_action('Select Options'))

        if self.default:
            prompt = '%s %s%s%s' % (prompt,
                styles_default_wrapper('['),
                styles_default_value(self.default),
                styles_default_wrapper(']'))

        return prompt

    def __call__(self, form, prefix=None, stylesheet=None):

        choices_count = len(self.choices)

        while 1:
            prompt = '%s' % (self.build_prompt(
                prefix=prefix,
                stylesheet=stylesheet))

            data = self.apply_default(input('%s ' % prompt))

            try:
                obj = next(x for x in self.choices if str(x[0]) == str(data))
            except StopIteration:
                continue

            if self.choices.index(obj) == (choices_count - 1):
                self.value = list(self.value)
                break

            if obj in self.value:
                self.value = self.value - frozenset([obj])
                continue

            try:
                self.process_data(data)
            except:
                continue

    def process_data(self, data):
        result = [x for x in self.choices if str(x[0]) == str(data)]

        if not result:
            raise ValueError

        self.value.add(result[0])


